#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (C) 2008-2018 Sergey Poznyakoff
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Wiki markup to plain text translator.

Classes:

TextWikiMarkup       -- Converts Wiki material to plain text.
TextWiktionaryMarkup -- Reserved for future use. Currently does the same as
                        TextWikiMarkup.

"""

from wikitrans.wikitoken import *
from wikitrans.wikimarkup import *
from wikitrans.wikins import wiki_ns_re, wiki_ns
import re
try:
    from urllib import quote as url_quote
except ImportError:
    from urllib.parse import quote as url_quote


class TextSeqNode(WikiSeqNode):
    def format(self):
        string = ""
        for x in self.content:
            if len(string) > 1 and not string[-1].isspace():
                string += ' '
            string += x.format()
        return string


class TextTextNode(WikiTextNode):
    def format(self):
        if isinstance(self.content, list):
            string = ""
            for s in self.content:
                if string:
                    if string.endswith("."):
                        string += "  "
                    else:
                        string += " "
                string += s
        else:
            string = self.content
        return string


class TextPreNode(WikiSeqNode):
    def format(self):
        string = ""
        for x in self.content:
            string += x.format()
        string += '\n'
        return string


class TextParaNode(WikiSeqNode):
    def format(self):
        string = ""
        for x in self.content:
            string += x.format()
        string = self.parser.fmtpara(string) + '\n\n'
        return string


class TextItNode(WikiSeqNode):
    def format(self):
        string = ""
        for x in self.content:
            s = x.format()
            if s:
                string += " " + s
        return "_" + string.lstrip(" ") + "_"


class TextBoldNode(WikiSeqNode):
    def format(self):
        string = ""
        for x in self.content:
            if string.endswith("."):
                string += "  "
            else:
                string += " "
            string += x.format()
        return string.upper()


class TextLinkNode(WikiSeqNode):
    def format(self):
        arg = self.content[0].format()
        if len(self.content) > 1:
            s = [x for x in map(lambda x: x.format(), self.content)]
            text = s[1]
        else:
            s = None
            text = None

        if s:
            if s[0] == 'disambigR' or s[0] == 'wikiquote':
                return ""
            if len(s) > 1 and s[1] == 'thumb':
                return ""
        (qual, sep, tgt) = arg.partition(':')
        if tgt != '':
            ns = self.parser.wiki_ns_name(qual)
            if ns:
                if ns == 'NS_IMAGE':
                    if not self.parser.show_urls:
                        return ""
                    text = "[%s: %s]" % (qual, text if text else arg)
                    tgt = "%s/%s/250px-%s" % (self.image_base,
                                              url_quote(tgt),
                                              url_quote(tgt))
                elif ns == 'NS_MEDIA':
                    text = "[%s]" % (qual)
                else:
                    tgt = self.parser.mktgt(tgt)
            elif self.type == 'LINK' and qual in self.parser.langtab:
                text = self.parser.langtab[qual] + ": " + tgt
                tgt = self.parser.mktgt(tgt, qual)
            else:
                tgt = self.parser.mktgt(tgt)
        else:
            tgt = self.parser.mktgt(arg)
        if self.parser.show_urls:
            return "%s (see %s) " % (text, tgt)
        elif not text or text == '':
            return arg
        else:
            return text


class TextTmplNode(TextLinkNode):
    def format(self):
        return '[' + super(TextTmplNode, self).format() + ']'


class TextBarNode(WikiNode):
    def format(self):
        w = self.parser.width
        if w < 5:
            w = 5
        return "\n" + ("-" * (w - 5)).center(w - 1) + "\n"


class TextHdrNode(WikiHdrNode):
    def format(self):
        return ("\n"
                + ("*" * self.level)
                + " "
                + self.content.format().lstrip(" ")
                + "\n\n")


class TextRefNode(WikiRefNode):
    def format(self):
        text = self.content.format()
        if text:
            return "%s (see %s) " % (text, self.ref)
        else:
            return "see " + self.ref


class TextEnvNode(WikiEnvNode):
    def format(self):
        type = self.envtype
        lev = self.level
        if lev > self.parser.width - 4:
            lev = 1
        string = ""
        n = 1
        for s in self.content:
            if not string.endswith("\n"):
                string += "\n"
            x = s.content.format()
            if type == "unnumbered":
                string += self.parser.indent(lev, "- " + x.lstrip(" "))
            elif type == "numbered":
                string += self.parser.indent(lev, "%d. %s" % (n, x))
                n += 1
            elif type == "defn":
                if s.subtype == 0:
                    string += self.parser.indent(lev-1, x)
                else:
                    string += self.parser.indent(lev+3, x)
            if not string.endswith("\n"):
                string += "\n"
        return string


class TextIndNode(WikiIndNode):
    def format(self):
        return (" " * self.level) + self.content.format() + '\n'


class TextTagNode(WikiTagNode):
    def format(self):
        if self.tag == 'code':
            self.parser.nested += 1
            s = self.content.format()
            self.parser.nested -= 1
        elif self.tag == 'ref':
            s = '[%d]' % (self.idx+1)
        elif self.tag == 'references':
            s = '\nReferences:\n'
            for ref in self.parser.references:
                s += ('[%d]. ' % (ref.idx+1))  + ref.content.format() + '\n'
        else:
            s = '<' + self.tag
            if self.args:
                s += ' ' + str(self.args)
            s += '>' + self.content.format() + '</' + self.tag + '>'
        return s


class TextWikiMarkup(WikiMarkup):
    """A Wiki markup to plain text translator.

    Usage:

      x = TextWikiMarkup(file="input.wiki")
      # Parse the input:
      x.parse()
      # Print it as plain text:
      print(str(x))

    """

    # Output width
    width = 78
    # Do not show references.
    show_urls = False
    # Provide a minimum markup
    markup = True

    # Number of current element in the environment
    num = 0

    # Array of footnote references
    references = []

    def __init__(self, *args, **keywords):
        """Create a TextWikiMarkup object.

        TextWikiMarkup([filename=FILE],[file=FD],[text=STRING],[lang=CODE],
                       [html_base=URL],[image_base=URL],[media_base=URL],
                       [width=N],[show_urls=False])

        Most arguments have the same meaning as in the WikiMarkup constructor.

        Class-specific arguments:

        width=N
          Limit output width to N columns. Default is 78.
        show_urls=False
          By default, the link URLs are displayed in parentheses next to the
          link text. If this argument is given, only the link text will be
          displayed.
        """

        super(TextWikiMarkup, self).__init__(*args, **keywords)
        if 'width' in keywords:
            self.width = keywords['width']
        if 'show_urls' in keywords:
            self.show_urls = keywords['show_urls']
        self.token_class['SEQ'] = TextSeqNode
        self.token_class['TEXT'] = TextTextNode
        self.token_class['PRE'] = TextPreNode
        self.token_class['PARA'] = TextParaNode
        self.token_class['SEQ'] = TextSeqNode
        self.token_class['IT'] = TextItNode
        self.token_class['BOLD'] = TextBoldNode
        self.token_class['LINK'] = TextLinkNode
        self.token_class['TMPL'] = TextTmplNode
        self.token_class['BAR'] = TextBarNode
        self.token_class['HDR'] = TextHdrNode
        self.token_class['REF'] = TextRefNode
        self.token_class['ENV'] = TextEnvNode
        self.token_class['IND'] = TextIndNode
        self.token_class['TAG'] = TextTagNode

    def wiki_ns_name(self, str):
        if str in wiki_ns[self.lang]:
            return wiki_ns[self.lang][str]
        elif str in wiki_ns_re[self.lang]:
            for elt in wiki_ns_re[self.lang][str]:
                if str.beginswith(elt[0]) and str.endswith(elt[1]):
                    return elt[2]
        return None

    def mktgt(self, tgt, lang = None):
        if not lang:
            lang = self.lang
        return self.html_base % { 'lang' : lang } + url_quote(tgt)

    def indent(self, lev, text):
        if text.find('\n') == -1:
            s = (" " * lev) + text
        else:
            s = ""
            for elt in text.split('\n'):
                if elt:
                    s += (" " * lev) + elt + '\n'
            if not text.endswith('\n'):
                s = s.rstrip('\n')
        return s

    def fmtpara(self, input):
        output = ""
        linebuf = ""
        length = 0
        for s in input.split():
            wlen = len(s)
            if len(linebuf) == 0:
                wsc = 0
            elif linebuf.endswith("."):
                wsc = 2
            else:
                wsc = 1
            if length + wsc + wlen > self.width:
                # FIXME: fill out linebuf
                output += linebuf + '\n'
                wsc = 0
                length = 0
                linebuf = ""
            linebuf += " " * wsc + s
            length += wsc + wlen
        return output + linebuf

    def __str__(self):
        str = ""
        for elt in self.tree:
            str += elt.format()
        return str


class TextWiktionaryMarkup(TextWikiMarkup):
    """A class for translating Wiktionary articles into plain text.

    Reserved for future use. Currently does the same as TextWikiMarkup.
    """

    html_base='http://%(lang)s.wiktionary.org/wiki/'



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (C) 2015-2018 Sergey Poznyakoff
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Wiki markup to Texinfo translator.

Classes:

TexiWikiMarkup       -- Converts Wiki material to Texinfo.

"""

from wikitrans.wikimarkup import *
from wikitrans.wikitoken import *
from wikitrans.wikins import wiki_ns_re, wiki_ns
import re
import urllib


class Acc(list):
    def prepend(self, x):
        self.insert(0, x)

    def is_empty(self):
        return len(self) == 0

    def clear(self):
        self = []

    def tail(self, n = 1):
        s = Acc()
        i = len(self)
        while i > 0 and n > 0:
            elt = self[i-1]
            l = len(elt)
            if l == 0:
                continue
            elif l > n:
                l = n
            s.prepend(elt[-n:])
            n -= l
            i -= 1
        return str(s)

    def trim(self, n):
        while len(self) and n > 0:
            elt = self.pop()
            l = len(elt)
            if l == 0:
                continue
            elif l > n:
                self += elt[0:-n]
                break
            n -= l

    def trimnl(self):
        if self.endswith('\n'):
            self.trim(1)

    def trimpara(self):
        if self.endswith('\n\n'):
            self.trim(2)

    def endswith(self, x):
        return self.tail(len(x)) == x

    def in_new_para(self):
        return self.is_empty() or self.endswith('\n\n')

    def __str__(self):
        return ''.join(self)


class TexiTextNode(WikiTextNode):
    def format(self):
        parser = self.parser
        if isinstance(self.content, list):
            for s in self.content:
                parser._print(s)
        else:
            parser._print(self.content)


class TexiTagNode(WikiTagNode):
    def format(self):
        parser = self.parser
        if self.tag in ['code', 'tt']:
            save = parser._begin_print()
            parser.nested += 1
            self.content.format()
            parser.nested -= 1
            s = parser._end_print(save)
            if self.isblock:
                parser._print('@example', nl=True, escape=False)
                parser._print(s, escape=False)
                parser._print('@end example\n', nl=True, escape=False)
            else:
                parser._print('@code{%s}' % s, escape=False)
        elif self.tag == 'div':
            if self.args and 'id' in self.args:
                parser._print("@anchor{%s}\n" % self.args['id'],
                              nl=True, escape=False)
            self.content.format()
        elif self.tag == 'ref':
            parser._print('@footnote{', escape=False);
            self.content.format();
            parser._print('}', escape=False)
        elif self.tag == 'references':
            pass
        else:
            parser._print('<' + self.tag)
            if self.args:
                parser._print(' ' + self.args)
            parser._print('>');
            self.content.format()
            parser._print('</' + self.tag + '>')


class TexiParaNode(WikiSeqNode):
    def format(self):
        parser = self.parser
        if not parser.acc.in_new_para():
            parser._print('\n', nl=True)
        for x in self.content:
            x.format()
        if not parser.acc.in_new_para():
            parser._print('\n', nl=True)


class TexiPreNode(WikiSeqNode):
    def format(self):
        parser = self.parser
        if not parser.nested:
            parser._print('@example\n', nl=True, escape=False)
        for x in self.content:
            x.format()
        if not parser.nested:
            parser._print('@end example\n', nl=True, escape=False)


class TexiFontNode(WikiSeqNode):
    def format(self):
        parser = self.parser
        comm = { 'IT': 'i',
                 'BOLD': 'b' }
        parser._print('@%s{' % comm[self.type], escape=False)
        for x in self.content:
            x.format()
        parser._print('}', escape=False)


class TexiHdrNode(WikiHdrNode):
    def format(self):
        parser = self.parser
        level = self.level
        # FIXME
        if level > len(parser.sectcomm[parser.sectioning_model]) - 1 - parser.sectioning_start:
            parser._print("@* ", nl=True, escape=False)
            self.content.format()
        else:
            parser._print(parser.sectcomm[parser.sectioning_model][level - parser.sectioning_start] + " ", nl=True, escape=False)
            self.content.format()
            parser._print(None, nl=True)
            if parser.sectcomm[parser.sectioning_model][0] == '@top':
                parser._print('@node ', nl=True, escape=False)
                self.content.format()
                parser._print('\n')
        parser._print(None, nl=True)


class TexiBarNode(WikiNode):
    def format(self):
        self.parser._print("\n-----\n")


class TexiIndNode(WikiIndNode):
    def format(self):
        parser = self.parser
        parser._print("@w{ }" * self.level, nl=True, escape=False)
        self.content.format()
        parser._print(None, nl=True)


class TexiEnvNode(WikiEnvNode):
    def format(self):
        parser = self.parser
        if self.envtype == 'unnumbered':
            parser._print('@itemize @bullet\n', nl=True, escape=False)
            for s in self.content:
                parser._print('@item ', nl=True, escape=False)
                s.content.format()
                parser._print(None, nl=True)
                parser._print('\n')
            parser._print('@end itemize\n', nl=True, escape=False)
        elif self.envtype == 'numbered':
            parser._print('@enumerate\n', nl=True, escape=False)
            for s in self.content:
                parser._print('@item ', nl=True, escape=False)
                s.content.format()
                parser._print(None, nl=True)
                parser._print('\n')
            parser._print('@end enumerate\n', nl=True, escape=False)
        elif self.envtype == 'defn':
            parser._print('@table @asis\n', nl=True, escape=False)
            for s in self.content:
                if s.subtype == 0:
                    parser._print('@item ', nl=True, escape=False)
                    s.content.format()
                    parser._print(None, nl=True)
                else:
                    s.content.format()
                    parser._print(None, nl=True)
                    parser._print('\n')
            parser._print('@end table\n', nl=True, escape=False)


class TexiLinkNode(WikiSeqNode):
    def format(self):
        parser = self.parser
        save = parser._begin_print()
        self.content[0].format()
        arg = parser._end_print()
        if len(self.content) > 1:
            s = []
            for x in self.content[0:2]:
                parser._begin_print()
                x.format()
                s.append(parser._end_print())
            text = s[1]
        else:
            s = None
            text = None

        parser._end_print(save)

        if s:
            if s[0] == 'disambigR' or s[0] == 'wikiquote':
                return
            if len(s) > 1 and s[1] == 'thumb':
                return

        (qual, sep, tgt) = arg.partition(':')
        if text:
            parser._print("@ref{%s,%s}" % (qual, text), escape=False)
        else:
            parser._print("@ref{%s}" % qual, escape=False)


class TexiRefNode(WikiRefNode):
    def format(self):
        parser = self.parser
        target = self.ref
        save = parser._begin_print()
        self.content.format()
        text = parser._end_print(save)
        if text and text != '':
            parser._print("@uref{%s,%s}" % (target, text), escape=False)
        else:
            parser._print("@uref{%s}" % target, escape=False)


class TexiWikiMarkup(WikiMarkup):
    """Wiki markup to Texinfo translator class.

    Usage:

      x = TexiWikiMarkup(file="input.wiki")
      # Parse the input:
      x.parse()
      # Print it as Texi:
      print(str(x))

    """

    nested = 0
    sectcomm = {
        'numbered': [
            '@top',
            '@chapter',
            '@section',
            '@subsection',
            '@subsubsection'
        ],
        'unnumbered': [
            '@top',
            '@unnumbered',
            '@unnumberedsec',
            '@unnumberedsubsec',
            '@unnumberedsubsubsec'
        ],
        'appendix': [
            '@top',
            '@appendix',
            '@appendixsec',
            '@appendixsubsec',
            '@appendixsubsubsec'
        ],
        'heading': [
            '@majorheading'
            '@chapheading',
            '@heading',
            '@subheading',
            '@subsubheading'
        ]
    }

    sectioning_model = 'numbered'
    sectioning_start = 0

    def __init__(self, *args, **keywords):
        """Create a TexiWikiMarkup object.

        TexiWikiMarkup([filename=FILE],[file=FD],[text=STRING],[lang=CODE],
                       [html_base=URL],[image_base=URL],[media_base=URL],
                       [sectioning_model=MODEL],[sectioning_start=N])

        For a discussion of generic arguments, see the constructor of
        the WikiMarkup class.

        Additional arguments:

        sectioning_model=MODEL
          Select the Texinfo sectioning model for the output document. Possible
          values are:

          'numbered'
             Top of document is marked with "@top". Headings ("=", "==",
             "===", etc) produce "@chapter", "@section", "@subsection", etc.
          'unnumbered'
             Unnumbered sectioning: "@top", "@unnumbered", "@unnumberedsec",
             "@unnumberedsubsec".
          'appendix'
             Sectioning suitable for appendix entries: "@top", "@appendix",
             "@appendixsec", "@appendixsubsec", etc.
          'heading'
             Use heading directives to reflect sectioning: "@majorheading",
             "@chapheading", "@heading", "@subheading", etc.
        sectioning_start=N
          Shift resulting heading level by N positions. For example, supposing
          "sectioning_model='numbered'", "== A ==" normally produces
          "@section A" on output. Now, if given "sectioning_start=1", this
          directive will produce "@subsection A" instead.
        """

        super(TexiWikiMarkup, self).__init__(*args, **keywords)

        self.token_class['TEXT'] = TexiTextNode
        self.token_class['TAG']  = TexiTagNode
        self.token_class['PARA'] = TexiParaNode
        self.token_class['PRE']  = TexiPreNode
        self.token_class['IT']   = TexiFontNode
        self.token_class['BOLD'] = TexiFontNode
        self.token_class['HDR']  = TexiHdrNode
        self.token_class['BAR']  = TexiBarNode
        self.token_class['IND']  = TexiIndNode
        self.token_class['ENV']  = TexiEnvNode
        self.token_class['LINK'] = TexiLinkNode
        self.token_class['REF']  = TexiRefNode

        if "sectioning_model" in keywords:
            val = keywords["sectioning_model"]
            if val in self.sectcomm:
                self.sectioning_model = val
            else:
                raise ValueError("Invalid value for sectioning model: %s" % val)
        if "sectioning_start" in keywords:
            val = keywords["sectioning_start"]
            if val < 0 or val > 4:
                raise ValueError("Invalid value for sectioning start: %s" % val)
            else:
                self.sectioning_start = val

    replchars = re.compile(r'([@{}])')
    acc = Acc()

    def _print(self, text, **kw):
        nl = kw.pop('nl', False)
        escape = kw.pop('escape', True)
        if nl and not self.acc.endswith('\n'):
            self.acc += '\n'
        if text:
            if escape:
                self.acc += self.replchars.sub(r'@\1', text)
            else:
                self.acc += text

    def _begin_print(self):
        s = self.acc
        self.acc = Acc()
        return s

    def _end_print(self, val = None):
        s = self.acc
        self.acc = val
        return str(s)

    def __str__(self):
        self._begin_print()
        for elt in self.tree:
            elt.format()
        self.acc.trimpara()
        return self._end_print()



import unittest
import sys
from os.path import dirname, abspath

sys.path.insert(0,dirname(abspath(__file__)))

def wiki_test_suite():
    loader = unittest.TestLoader()
    return loader.loadTestFromModule()




#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (C) 2008-2018 Sergey Poznyakoff
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Wiki markup parser.

This module provides two class:

WikiMarkupParser:
   An abstract parser class, which serves as a base class for all markup
   classes in this package.

WikiMarkup
   A subclass of the above, providing basic input method.

"""

from __future__ import print_function
import sys
import re
from types import *
from wikitrans.wikitoken import *

__all__ = [ "WikiMarkupParser", "WikiMarkup",
            "TagAttributes", "TagAttributeSyntaxError" ]


class UnexpectedTokenError(Exception):
    def __init__(self, value):
        self.value = value


class TagAttributeSyntaxError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class TagAttributes(object):
    """A dictionary-like collection of tag attributes.

    Example:

      attr = TagAttributes('href="foo" length=2')
      if 'href' in attr:
          print(x['href'])   # returns "foo"
      for a in attr:
          ...
    """

    attrstart = re.compile("^(?P<attr>[a-zA-Z0-9_-]+)(?P<eq>=\")?")
    valseg = re.compile("^[^\\\"]+")
    tab = {}
    printable = None
    def __init__(self, string):
        if not string:
            self.printable = ''
            return
        self.printable = string
        s = string
        self.tab = {}
        while s != '':
            s = s.strip()
            m = self.attrstart.match(s)
            if m:
                name = m.group('attr')
                val = ''
                s = s[m.end(0):]
                if m.group('eq'):
                    while 1:
                        m = self.valseg.match(s)
                        val += m.group(0)
                        s = s[m.end(0):]
                        if s[0] == '\\':
                            val += s[1]
                            s += 2
                        elif s[0] == '"':
                            s = s[1:]
                            break
                else:
                    val = 1
                self.tab[name] = val
            else:
                raise TagAttributeSyntaxError(s)

    def __len__(self):
        return len(self.tab)

    def __getitem__(self, key):
        return self.tab[key]

    def __contains__(self, key):
        return key in self.tab

    def __iter__(self):
        for key in self.tab:
            yield(key)

    def has_key(self, key):
        return self.__contains__(key)

    def __setitem__(self, key, value):
        self.tab[key] = value

    def __delitem__(self, key):
        del self.tab[key]

    def __str__(self):
        return self.printable

    def __repr__(self):
        return self.printable


class WikiMarkupParser(object):
    """Parser for Wiki markup language.

    Given input in Wiki markup language creates an abstract parse tree for it.
    This is a base class for actual parsers. The subclasses must provide the
    input method.

    Public methods:

    parse()  --  parse the input.

    Abstract methods (must be overridden by the subclass):

    input()  --  returns next physical line from the input material.

    Public attributes:

    Input:
    debug_level -- debug verbosity level (0 - no debug info, 100 - excessively
                   copious debug messages). Default is 0.
    strict      -- if True, parser will throw exception upon encountering
                   invalid markup tag (mostly for future use)

    Output:
    tree     --  constructed parse tree (a subclass of WikiNode)

    """

    delim = re.compile("^==+[ \t]*|[ \t]*==+[ \t]*$|(^----$)|^\\*+|^#+|^[;:]+|(\\[\\[)|\\[|(\\{\\{)|(\\]\\])|\\]|(\\}\\})|\\||(\\'\\'\\'?)|<")
    otag = re.compile("<(?P<tag>[a-zA-Z0-9_]+)(?:\s+(?P<args>[^/][^>]+))?\s*(?P<closed>/)?>")
    ctag = re.compile("</(?P<tag>[a-zA-Z0-9_]+)\s*>")
    refstart = re.compile("^https?://")

    close_delim = {
        '[': ']',
        '[[': ']]',
        '{{': '}}'
    }

    # Environment types:
    envtypes = { "*": [ "unnumbered", 0 ],
                 "#": [ "numbered", 0 ],
                 ";": [ "defn", 0 ],
                 ":": [ "defn", 1 ]
    }

    toklist = None
    tokind = 0
    newline = 0
    tree = None

    tags = [ 'code', 'nowiki', 'tt', 'div', 'ref', 'references' ]

    debug_level = 0
    strict = False

    def dprint(self, lev, fmt, *argv):
        """If current debug level is greater than or equal to lev, print *argv
        according to format.
        """
        if self.debug_level >= lev:
            for l in (fmt % argv).split('\n'):
                print("[DEBUG] %s" % l)

    inline_delims = [ "''", "'''", "[", "]", "[[", "]]", "{{", "}}", "|" ]

    token_class = {}

    def _new_node(self, **kwarg):
        return self.token_class[kwarg['type']](self, **kwarg)

    def tokread(self):
        """Read next token from the input. Return it as a subclass of WikiNode."""
        line = None
        pos = 0
        while 1:
            if (not line or pos == len(line)):
                try:
                    line = self.input()
                    pos = 0
                except StopIteration:
                    line = u''

            if not line or line == "":
                yield(self._new_node(type='NIL'))
                break

            if line == '\n':
                yield(self._new_node(type='NL'))
                line = None
                continue

            self.dprint(100, "LINE: %s", line[pos:])
            m = self.delim.search(line, pos)

            if m:
                if (pos < m.start(0)):
                    yield(self._new_node(type='TEXT',
                                         content=line[pos:m.start(0)]))
                pos = m.start(0)
                t = None

                if line[m.start(0)] == '<':
                    m = self.otag.match(line, pos)
                    if m:
                        pos = m.end(0)
                        if m.group('tag') == 'nowiki':
                            if not m.group('closed'):
                                while 1:
                                    try:
                                        m = self.ctag.search(line, pos)
                                        if m and m.group('tag') == 'nowiki':
                                            yield(self._new_node(type='TEXT',
                                                                 content=line[pos:m.start(0)] ))
                                            pos = m.end(0)
                                            break

                                        yield(self._new_node(type='TEXT',
                                                             content=line[pos:]))

                                        line = self.input()
                                        pos = 0
                                    except StopIteration:
                                        break
                            continue
                        elif m.group('tag') in self.tags:
                            try:
                                yield(self._new_node(type='OTAG',
                                                  tag=m.group('tag'),
                                                  isblock=(line[pos] == '\n'),
                                                  args=TagAttributes(m.group('args'))))
                                if m.group('closed'):
                                    yield(self._new_node(type='CTAG',
                                                         tag=m.group('tag')))
                            except TagAttributeSyntaxError:
                                yield(self._new_node(type='TEXT',
                                                     content=m.group(0)))
                            continue
                        else:
                            yield(self._new_node(type='TEXT',
                                                 content=m.group(0)))
                            continue
                    else:
                        m = self.ctag.match(line, pos)
                        if m:
                            if m.group('tag') in self.tags:
                                yield(self._new_node(type='CTAG',
                                                     tag=m.group('tag')))
                            else:
                                yield(self._new_node(type='TEXT',
                                                     content=m.group(0)))
                            pos = m.end(0)
                            continue
                        else:
                            yield(self._new_node(type='TEXT',
                                                 content=line[pos:pos+1]))
                            pos += 1
                            continue
                else:
                    pos = m.end(0)
                    content = m.group(0)
                    if content[0] in self.envtypes:
                        node = self._new_node(type='DELIM',
                                              content=content,
                                              isblock=True,
                                              continuation=pos < len(line) and line[pos] == ":")
                        if node.continuation:
                            node.content += node.content[0]
                            pos += 1

                        yield(node)

                        while pos < len(line) and line[pos] in [' ', '\t']:
                            pos += 1
                    else:
                        yield(self._new_node(type='DELIM',
                                             isblock=(content.strip() not in self.inline_delims),
                                             content=content.strip()))
                    continue

            if line:
                if line[-1] == '\n':
                    if line[pos:-1] != '':
                        yield(self._new_node(type='TEXT', content=line[pos:-1]))
                    yield(self._new_node(type='NL'))
                else:
                    yield(self._new_node(type='TEXT', content=line[pos:]))
                line = None

    def input(self):
        """Return next physical line from the input.

        This method must be overridden by the subclass.
        """
        return None

    def swaptkn(self, i, j):
        """Swap tokens at indices i and j in toklist."""
        self.dprint(80, "SWAPPING %s <-> %s", i, j)
        x = self.toklist[i]
        self.toklist[i] = self.toklist[j]
        self.toklist[j] = x

    def tokenize(self):
        """Tokenize the input.

        Read tokens from the input (supplied by the input() method). Place the
        obtained tokens in the toklist array.
        """
        self.toklist = []
        for tok in self.tokread():
            self.dprint(100, "TOK: %s", tok)
            self.toklist.append(tok)
        # Determine and fix up the ordering of bold and italic markers
        # There are three possible cases:
        #
        # 1a. '''a b ''c'' d'''
        # 1b. ''a b '''c''' d''
        #
        # 2a. '''''a b'' c d'''
        # 2b. '''''a b''' c d''
        #
        # 3a. '''a b ''c d'''''
        # 3b. ''a b '''c d'''''
        stack = []
        for i in range(0, len(self.toklist)):
            if (self.toklist[i].type == 'DELIM'
                and (self.toklist[i].content == "''"
                     or self.toklist[i].content == "'''")):
                if len(stack) > 0:
                    if self.toklist[stack[-1]].content == self.toklist[i].content:
                        # Case 1: just pop the matching delimiter off the stack
                        stack.pop()
                    elif len(stack) == 2 and stack[-2] + 1 == stack[-1]:
                        # Case 2: swap delimiters saved on stack ...
                        self.swaptkn(stack[-2], stack[-1])
                        #         and pop off the matching one
                        stack.pop()
                    elif (i < len(self.toklist)
                          and self.toklist[i+1].type == 'DELIM'
                          and self.toklist[stack[-1]].content
                                      == self.toklist[i+1].content):
                        # Case 3: swap current and next tokens
                        self.swaptkn(i, i+1)
                        #         and pop off the matching one
                        stack.pop()
                    else:
                        # Push the token on stack
                        stack.append(i)
                else:
                    # Push the token on stack
                    stack.append(i)
        # Redefine all non-matched tokens as TEXT
        for i in stack:
            # FIXME: How to convert node to TEXT?
            self.toklist[i] = self._new_node(type='TEXT',
                                             content=str(self.toklist[i]))

    mark = []

    def push_mark(self):
        """Save the current token index on stack."""
        self.mark.append(self.tokind)

    def pop_mark(self):
        """Restore the token index from top of stack."""
        self.tokind = self.mark.pop()

    def clear_mark(self):
        """Forget the last mark."""
        self.mark.pop()

    def lookahead(self, off=0):
        """Peek a token at index (tokind+off)."""
        tok = self.toklist[self.tokind+off]
        self.dprint(20, "lookahead(%s): %s", off, tok)
        return tok

    def setkn(self, val):
        """Store token val at the current token index."""
        self.toklist[self.tokind] = val

    def getkn(self):
        """Get next token from the toklist. Advance tokind."""
        self.newline = self.tokind == 0 or self.toklist[self.tokind-1].type == 'NL'
        if self.tokind == len(self.toklist):
            return self._new_node(type='NIL')
        tok = self.toklist[self.tokind]
        self.tokind = self.tokind + 1
        self.dprint(20, "getkn: %s", tok)
        return tok

    def ungetkn(self, tok=None):
        """Unget the last read token.

        Decrease the tokind by one, so the last read token will be read again.
        If optional argument is supplied and is not None, store it in the toklist
        in place of the current token.
        """
        self.tokind = self.tokind - 1
        self.newline = self.tokind == 0 or self.toklist[self.tokind-1].type == 'NL'
        if tok:
             self.toklist[self.tokind] = tok
        self.dprint(20, "ungetkn: %s", tok)
        return self.toklist[self.tokind]

    def fixuptkn(self, tok):
        """Replace the recently read token by tok."""
        if self.tokind == 0:
            raise IndexError('WikiMarkupParser.fixuptkn called at start of input')
        self.toklist[self.tokind-1] = tok
        return tok

    def dump(self, tree, file=sys.stdout):
        """Dump the tree to file, node by node."""
        for node in tree:
            file.write(str(node))
            file.write('\n')

    def is_block_end(self, tok):
        """Return True if tok ends a block environment."""
        if tok.type == 'NIL':
            return True
        elif tok.type == 'NL':
            if self.lookahead().type == 'NIL':
                return True
            elif self.lookahead().type == 'NL':
                self.getkn()
                return True
        elif tok.type in ['DELIM', 'CTAG', 'TAG']:
            if tok.isblock:
                self.ungetkn(tok)
                return True
        return False

    def parse_para(self, tok):
        """Read paragraph starting at tok."""
        self.dprint(80, "ENTER parse_para: %s", tok)

        acc = { 'seq': [],
                'textlist': [] }

        def flush():
            if acc['textlist']:
                acc['seq'].append(self._new_node(type='TEXT',
                                                 content=''.join(acc['textlist'])))
                acc['textlist'] = []

        if (isinstance(tok, WikiContentNode)
            and isinstance(tok.content, str)
            and re.match("^[ \t]", tok.content)):
            type = 'PRE'
            rx = re.compile("^\S")
        else:
            type = 'PARA'
            rx = re.compile("^[ \t]")

        while not self.is_block_end(tok):
            if tok.type == 'TEXT':
                if rx and self.newline and rx.match(tok.content):
                    self.ungetkn()
                    break
                acc['textlist'].append(tok.content)
            elif tok.type == 'NL':
                acc['textlist'].append('\n')
            elif tok.type == 'OTAG':
                flush()
                acc['seq'].append(self.parse_tag(tok))
            elif tok.type == 'DELIM':
                flush()
                acc['seq'].append(self.parse_inline_delim(tok))
            else:
                if self.strict:
                    raise UnexpectedTokenError(tok)
                # FIXME: Another possible variant of handling this case is to
                # convert tok to TEXT node and append it to acc['seq']
            tok = self.getkn()
        flush()
        if acc['seq']:
            tok = self._new_node(type=type, content=acc['seq'])
        else:
            tok = None
        self.dprint(80, "LEAVE parse_para=%s", tok)
        return tok

    def parse_block_delim(self, tok):
        """Parse block environment starting at tok."""
        self.dprint(80, "ENTER parse_block_delim")
        assert(tok.type == 'DELIM')
        if tok.content == "----":
            node = self._new_node(type = 'BAR')
        elif tok.content[0:2] == "==":
            node = self.parse_header(tok)
            if not node:
                tok = self.ungetkn(self._new_node(type='TEXT',
                                                  content=tok.content))
        elif tok.content[0] in self.envtypes:
            node = None
            if tok.content[0] == ':':
                t = self.lookahead(-2)
                if not (t.type == 'DELIM' and t.content == ';'):
                    node = self.parse_indent(tok)
            if not node:
                node = self.parse_env(tok)
        else:
            self.ungetkn(tok)
            node = None
        self.dprint(80, "LEAVE parse_block_delim=%s", node)
        return node

    def parse_line(self):
        """Parse the input line."""
        self.dprint(80, "ENTER parse_line")
        list = []
        while True:
            tok = self.getkn()
            if tok.type == 'NL' or tok.type == 'NIL':
                break
            elif tok.type == 'TEXT':
                list.append(tok)
            elif tok.type == 'DELIM':
                if tok.isblock:
                    tok = self._new_node(type = 'TEXT', content = tok.content)
                    self.fixuptkn(tok)
                    list.append(tok)
                elif tok.content[0] == ":":
                    # FIXME
                    list.append(self.parse_indent(tok))
                    break
                else:
                    x = self.parse_inline_delim(tok)
                    if x:
                        list.append(x)
                    else:
                        list.append(self.fixuptkn(self._new_node(type = 'TEXT',
                                                                 content = tok.content)))
            elif tok.type == 'OTAG':
                if tok.isblock:
                    self.ungetkn()
                    break
                list.append(self.parse_tag(tok))
            else:
                list.append(tok)
        ret = self._new_node(type='SEQ', content=list)
        self.dprint(80, "LEAVE parse_line=%s", ret)
        return ret

    def parse_indent(self, tok):
        """Parse indented block starting at tok."""
        lev = len(tok.content)
        self.dprint(80, "ENTER parse_indent(%s)", lev)
        x = self._new_node(type='IND', level=lev, content=self.parse_line())
        self.dprint(80, "LEAVE parse_indent=%s", x)
        return x

    def parse_fontmod(self, delim, what):
        """Parse font modification directive (bold or italics).

        Arguments:

        delim    --  starting delimiter ("''" or "'''")
        what     --  'IT' or 'BOLD'
        """
        self.dprint(80, "ENTER parse_fontmod(%s,%s), tok %s",
                    delim, what, self.lookahead())
        seq = []
        text = ''
        while True:
            tok = self.getkn()
            if tok.type == 'TEXT':
                text += tok.content
            elif self.is_block_end(tok):
                self.dprint(80, "LEAVE parse_fontmod=%s", "None")
                return None
            elif tok.type == 'DELIM':
#                self.dprint(80, "got %s, want %s", tok.content, delim)
                if tok.content == delim:
                    break
                else:
                    if text:
                        seq.append(self._new_node(type='TEXT', content=text))
                        text = ''
                    x = self.parse_inline_delim(tok)
                    if x:
                        seq.append(x)
                    else:
                        self.dprint(80, "LEAVE parse_fontmod=%s", "None")
                        return None
            elif tok.type == 'NL':
                seq.append(self._new_node(type='TEXT', content='\n'))
            else:
                self.dprint(80, "LEAVE parse_fontmod=None")
                return None
        if text:
            seq.append(self._new_node(type='TEXT', content=text))
        res = self._new_node(type=what, content=seq)
        self.dprint(80, "LEAVE parse_fontmod=%s", res)
        return res

    def parse_ref(self):
        """Parse a reference block ([...])"""
        self.dprint(80, "ENTER parse_ref")
        tok = self.getkn()
        if not (tok.type == 'TEXT' and self.refstart.match(tok.content)):
            self.dprint(80, "LEAVE parse_ref=None")
            return None

        seq = []
        (ref, sep, text) = tok.content.partition(' ')
        if text:
            seq.insert(0, self._new_node(type='TEXT', content=text))

        while True:
            tok = self.getkn()
            if tok.type == 'NIL':
                self.dprint(80, "LEAVE parse_ref=None")
                return None
            elif self.is_block_end(tok):
                self.dprint(80, "LEAVE parse_ref=None")
                return None
            elif tok.type == 'DELIM':
                if tok.content == ']':
                    break
                else:
                    tok = self.parse_inline_delim(tok)
                    if tok:
                        seq.append(tok)
                    else:
                        self.dprint(80, "LEAVE parse_ref=None")
                        return None
            elif tok.type == 'OTAG':
                list.append(self.parse_tag(tok))
            else:
                seq.append(tok)

        ret = self._new_node(type='REF', ref=ref,
                             content=self._new_node(type='SEQ', content=seq))
        self.dprint(80, "LEAVE parse_ref= %s", ret)
        return ret

    def parse_link(self, type, delim):
        """Parse an external link ([[...]]).

        In this implementation, it is also used to parse template
        references ({{...}}).

        Arguments:

        type   -- 'LINK' or 'TMPL'
        delim  -- expected closing delimiter.
        """
        self.dprint(80, "ENTER parse_link(%s,%s)", type, delim)
        subtree = []
        list = []
        while True:
            tok = self.getkn()
            if tok.type == 'NIL':
                self.dprint(80, "LEAVE parse_link=None [EOF]")
                return None
            if tok.type == 'DELIM':
                if tok.content == delim:
                    if list:
                        subtree.append(self._new_node(type='SEQ',
                                                      content=list))
                    break
                elif tok.content == "|":
                    if len(list) > 1:
                        subtree.append(self._new_node(type='SEQ',
                                                      content=list))
                    elif list:
                        subtree.append(list[0])
                    list = []
                else:
                    x = self.parse_inline_delim(tok)
                    if x:
                        list.append(x)
                    else:
                        self.dprint(80, "LEAVE parse_link=None [bad inline]")
                        return None
            elif tok.type == 'TEXT':
                list.append(tok)
            else:
                self.dprint(80, "LEAVE parse_link=None [unexpected token]")
                return None
        ret = self._new_node(type=type, content=subtree)
        self.dprint(80, "LEAVE parse_link=%s", ret)
        return ret

    def parse_inline_delim(self, tok):
        """Parse an inline block."""
        self.dprint(80, "ENTER parse_inline_delim")
        assert(tok.type == 'DELIM')
        self.push_mark()
        if tok.content == "''":
            x = self.parse_fontmod(tok.content, 'IT')
        elif tok.content == "'''":
            x = self.parse_fontmod(tok.content, 'BOLD')
        elif tok.content == "[":
            x = self.parse_ref()
        elif tok.content == "[[":
            x = self.parse_link('LINK', "]]")
        elif tok.content == "{{":
            x = self.parse_link('TMPL', "}}")
        else:
            x = None

        if x:
            self.clear_mark()
        else:
            self.dprint(80, "BEGIN DELIMITER RECOVERY: %s", tok)
            self.pop_mark()
            x = self.fixuptkn(self._new_node(type='TEXT', content=tok.content))
            od = tok.content
            if od in self.close_delim:
                cd = self.close_delim[od]
                lev = 0
                for i, tok in enumerate(self.toklist[self.tokind+1:]):
                    if tok.type == 'NIL':
                        break
                    elif tok.type == 'DELIM':
                        if tok.content == od:
                            lev += 1
                        elif tok.content == cd:
                            if lev == 0:
                                tok = self._new_node(type='TEXT',
                                                     content=tok.content)
                                self.toklist[self.tokind+1+i] = tok
                            lev -= 1
                            break
            self.dprint(80, "END DELIMITER RECOVERY: %s", tok)

        self.dprint(80, "LEAVE parse_inline_delim=%s", x)
        return x

    def parse_tag(self, tag):
        """Parse an xml-like tag (such as, e.g. "<tt>...</tt>")."""
        self.dprint(80, "ENTER parse_tag")
        list = []
        self.push_mark()
        while True:
            tok = self.getkn()
            if tok.type == 'NIL':
                self.pop_mark()
                s = '<' + tag.tag
                if tag.args:
                    s += ' ' + str(tag.args)
                s += '>'
                node = self._new_node(type='TEXT', content=s)
                if tag.content:
                    self.tree[self.tokind:self.tokind] = tag.content
                self.dprint(80, "LEAVE parse_tag = %s (tree modified)", node)
                return node
            elif tok.type == 'DELIM':
                if tok.isblock:
                    tok = self.parse_block_delim(tok)
                else:
                    tok = self.parse_inline_delim(tok)
                if not tok:
                    tok = self.getkn()
            elif tok.type == 'CTAG':
                if tag.tag == tok.tag:
                    break
                s = '</' + tag.tag + '>'
                tok = self.fixuptkn(self._new_node(type='TEXT', content=s))
            elif tok.type == 'NL':
                tok = self._new_node(type = 'TEXT', content = '\n')
            list.append(tok)
        self.clear_mark()
        ret = self._new_node(type = 'TAG',
                             tag  = tag.tag,
                             args = tag.args,
                             isblock = tag.isblock,
                             content = self._new_node(type = 'SEQ',
                                                      content = list))
        self.dprint(80, "LEAVE parse_tag = %s", ret)
        return ret

    def parse_env(self, tok):
        """Parse a block environment (numbered, unnumbered, or definition list)."""
        type = self.envtypes[tok.content[0]][0]
        lev = len(tok.content)
        self.dprint(80, "ENTER parse_env(%s,%s)", type, lev)
        list = []
        while True:
            if (tok.type == 'DELIM'
                and tok.content[0] in self.envtypes
                and type == self.envtypes[tok.content[0]][0]):
                if len(tok.content) < lev:
                    self.ungetkn()
                    break
                elif len(tok.content) > lev:
                    elt = self.parse_env(tok)
                else:
                    elt = self.parse_line()
                    if not tok.continuation:
                        list.append(self._new_node(type='ELT',
                                        subtype=self.envtypes[tok.content[0]][1],
                                        content=elt))
                        tok = self.getkn()
                        continue

                if list:
                    if list[-1].content.type != 'SEQ':
                        x = list[-1].content.content
                        # FIXME:
                        list[-1].content = self._new_node(type='SEQ', content=[x])
                    list[-1].content.content.append(elt)
            else:
                self.ungetkn()
                break

            tok = self.getkn()
        ret = self._new_node(type='ENV',
                             envtype=type,
                             level=lev,
                             content=list)
        self.dprint(80, "LEAVE parse_env=%s", ret)
        return ret

    def parse_header(self, tok):
        """Parse a Wiki header."""
        self.dprint(80, "ENTER parse_header")
        self.push_mark()
        list = []
        delim = tok.content
        while True:
            tok = self.getkn()
            if tok.type == 'NL':
                self.pop_mark()
                self.dprint(80, "LEAVE parse_header=None")
                return None
            elif tok.type == 'TEXT':
                list.append(tok)
            elif tok.type == 'DELIM':
                if tok.content == delim:
                    if self.lookahead().type == 'NL':
                        self.getkn()
                        if self.lookahead().type == 'NL':
                            self.getkn()
                        break
                    else:
                        self.pop_mark()
                        self.dprint(80, "LEAVE parse_header=None")
                        return None
                elif tok.isblock:
                    self.pop_mark()
                    self.dprint(80, "LEAVE parse_header=None")
                    return None
                else:
                    list.append(self.parse_inline_delim(tok))
            elif tok.type == 'OTAG':
                if tok.isblock:
                    self.pop_mark()
                    self.dprint(80, "LEAVE parse_header=None")
                    return None
                list.append(self.parse_tag(tok))
        self.clear_mark()
        ret = self._new_node(type='HDR',
                             level=len(delim),
                             content=self._new_node(type='SEQ', content=list))
        self.dprint(80, "LEAVE parse_header=%s", ret)
        return ret

    def parse_block(self):
        """Parse next block: newline, delimiter, tag, or paragraph."""
        tok = self.getkn()
        while tok.type == 'NL':
            tok = self.getkn()
        if tok == None or tok.type == 'NIL':
            return None
        elif tok.type == 'DELIM':
            tok = self.parse_block_delim(tok)
            if tok:
                return tok
            else:
                tok = self.getkn()
        elif tok.type == 'OTAG' and tok.isblock:
            return self.parse_tag(tok)

        return self.parse_para(tok)

    def parse(self):
        """Parse Wiki material supplied by the input() method.

        Store the resulting abstract parsing tree in the tree attribute.
        """
        if not self.toklist:
            self.tokenize()
        if self.debug_level >= 90:
            print("TOKEN DUMP BEGIN")
            self.dump(self.toklist)
            print("TOKEN DUMP END")
        self.tokind = 0
        self.tree = []
        while 1:
            subtree = self.parse_block()
            if subtree == None:
                break
            self.tree.append(subtree)
        if self.debug_level >= 70:
            print("TREE DUMP BEGIN")
            self.dump(self.tree)
            print("TREE DUMP END")

    def __str__(self):
        return str(self.tree)


class WikiMarkup(WikiMarkupParser):
    """
    A derived parser class that supplies a basic input method.

    Three types of inputs are available:

    1. filename=<file>
    The file <file> is opened and used for input.
    2. file=<file>
    The already opened file <file> is used for input.
    3. text=<string>
    Input is taken from <string>, line by line.

    Usage:

    obj = WikiMarkup(arg=val)
    obj.parse
    ... Do whatever you need with obj.tree ...

    """

    file = None
    text = None
    lang = 'en'
    html_base = 'http://%(lang)s.wikipedia.org/wiki/'
    image_base = 'http://upload.wikimedia.org/wikipedia/commons/thumb/a/bf'
    media_base = 'http://www.mediawiki.org/xml/export-0.3'

    def __init__(self, *args, **keywords):
        """Create a WikiMarkup object.

        Arguments:

        filename=FILE
          Read Wiki material from the file named FILE.
        file=FD
          Read Wiki material from file object FD.
        text=STRING
          Read Wiki material from STRING.
        lang=CODE
          Specifies source language. Default is 'en'. This variable can be
          referred to as '%(lang)s' in the keyword arguments below.
        html_base=URL
          Base URL for cross-references. Default is
              'http://%(lang)s.wikipedia.org/wiki/'
        image_base=URL
          Base URL for images. Default is
              'http://upload.wikimedia.org/wikipedia/commons/thumb/a/bf'
        media_base=URL
          Base URL for media files. Default is
              'http://www.mediawiki.org/xml/export-0.3'

        debug_level=INT
          debug verbosity level (0 - no debug info, 100 - excessively
                                 copious debug messages). Default is 0.
        strict=BOOL
          Strict parsing mode. Throw exceptions on syntax errors. Default
          is False.
        """
        self.token_class = {
            'NIL':   WikiNode,
            'NL':    WikiNode,
            'OTAG':  WikiTagNode,
            'CTAG':  WikiTagNode,
            'TAG':   WikiTagNode,
            'DELIM': WikiDelimNode,
            'TEXT':  WikiTextNode,
            'PRE':   WikiContentNode,
            'PARA':  WikiSeqNode,
            'BAR':   WikiNode,
            'SEQ':   WikiSeqNode,
            'IND':   WikiIndNode,
            'REF':   WikiRefNode,
            'TMPL':  WikiSeqNode,
            'IT':    WikiSeqNode,
            'BOLD':  WikiSeqNode,
            'ELT':   WikiEltNode,
            'ENV':   WikiEnvNode,
            'LINK':  WikiSeqNode,
            'HDR':   WikiHdrNode
        }

        for kw in keywords:
            if kw == 'file':
                self.file = keywords[kw]
            elif kw == 'filename':
                self.file = open(keywords[kw])
            elif kw == 'text':
                if sys.version_info[0] > 2:
                    self.text = keywords[kw].decode('utf-8').split("\n")
                else:
                    self.text = keywords[kw].split("\n")
            elif kw == 'lang':
                self.lang = keywords[kw]
            elif kw == 'html_base':
                self.html_base = keywords[kw]
            elif kw == 'image_base':
                self.image_base = keywords[kw]
            elif kw == 'media_base':
                self.media_base = keywords[kw]
            elif kw == 'strict':
                self.strict = keywords[kw]
            elif kw == 'debug_level':
                self.debug_level = keywords[kw]

    def __del__(self):
        if self.file:
            self.file.close()

    def input(self):
        if self.file:
            return self.file.readline()
        elif self.text:
            return self.text.pop(0) + '\n'
        else:
            return None

    # ISO 639
    langtab = {
        "aa": "Afar",            # Afar
        "ab": "Аҧсуа",           # Abkhazian
        "ae": None,              # Avestan
        "af": "Afrikaans",       # Afrikaans
        "ak": "Akana",           # Akan
        "als": "Alemannisch",
        "am": "አማርኛ",            # Amharic
        "an": "Aragonés",        # Aragonese
        "ang": "Englisc",
        "ar": "العربية" ,        # Arabic
        "arc": "ܐܪܡܝܐ",
        "as": "অসমীয়া",         # Assamese
        "ast": "Asturian",
        "av": "Авар",            # Avaric
        "ay": "Aymara",           # Aymara
        "az": "Azərbaycan" ,     # Azerbaijani

        "ba": "Башҡорт",         # Bashkir
        "bar": "Boarisch",
        "bat-smg": "Žemaitėška",
        "bcl": "Bikol",
        "be": "Беларуская",      # Byelorussian; Belarusian
        "be-x-old": "Беларуская (тарашкевіца)",
        "bg": "Български",       # Bulgarian
        "bh": "भोजपुरी",  # Bihari
        "bi": "Bislama",         # Bislama
        "bm": "Bamanankan",      # Bambara
        "bn": "বাংলা" ,          # Bengali; Bangla
        "bo": "བོད་སྐད",         # Tibetan
        "bpy": "ইমার ঠার/বিষ্ণুপ্রিয়া মণিপুরী" ,
        "br": "Brezhoneg" ,      # Breton
        "bs": "Bosanski" ,       # Bosnian
        "bug": "Basa Ugi",
        "bxr": "Буряад",

        "ca": "Català" ,         # Catalan
        "cbk-zam": "Chavacano de Zamboanga",
        "cdo": "Mìng-dĕ̤ng-ngṳ̄",
        "cho": "Choctaw",
        "ce": "Нохчийн",         # Chechen
        "ceb": "Sinugboanong Binisaya" , # Cebuano
        "ch": "Chamor",          # Chamorro
        "chr": "ᏣᎳᎩ",
        "chy": "Tsetsêhestâhese",
        "co": "Cors",            # Corsican
        "cr": "Nehiyaw",         # Cree
        "crh": "Qırımtatarca",
        "cs": "Česky" ,          # Czech
        "csb": "Kaszëbsczi",
        "c": "Словѣньскъ",       # Church Slavic
        "cv": "Чăваш",           # Chuvash
        "cy": "Cymraeg" ,        # Welsh

        "da": "Dansk" ,          # Danish
        "de": "Deutsch" ,        # German
        "diq": "Zazaki",         # Dimli (Southern Zazaki)
        "dsb": "Dolnoserbski",
        "dv": "ދިވެހިބަސް",      # Divehi
        "dz": "ཇོང་ཁ",           # Dzongkha; Bhutani

        "ee": "Eʋegbe",          # Ewe
        "el": "Ελληνικά" ,       # Greek
        "eml": "Emiliàn e rumagnòl",
        "en": "English" ,        # English
        "eo": "Esperanto" ,
        "es": "Español" ,        # Spanish
        "et": "Eesti" ,          # Estonian
        "eu": "Euskara" ,         # Basque
        "ext": "Estremeñ",

        "fa": "فارسی" ,          # Persian
        "ff": "Fulfulde",        # Fulah
        "fi": "Suomi" ,          # Finnish
        "fiu-vro": "Võro",
        "fj": "Na Vosa Vakaviti",# Fijian; Fiji
        "fo": "Føroyskt" ,       # Faroese
        "fr": "Français" ,       # French
        "frp": "Arpitan",
        "fur": "Furlan",
        "fy": "Frysk",           # Frisian

        "ga": "Gaeilge",         # Irish
        "gan": "贛語 (Gànyŭ)",
        "gd": "Gàidhlig",        # Scots; Gaelic
        "gl": "Gallego" ,        # Gallegan; Galician
        "glk": "گیلکی",
        "got": "𐌲𐌿𐍄𐌹𐍃𐌺𐍉𐍂𐌰𐌶𐌳𐌰",
        "gn": "Avañe'ẽ",         # Guarani
        "g": "ગુજરાતી",          # Gujarati
        "gv": "Gaelg",           # Manx

        "ha": "هَوُسَ",          # Hausa
        "hak": "Hak-kâ-fa / 客家話",
        "haw": "Hawai`i",
        "he": "עברית" ,          # Hebrew (formerly iw)
        "hi": "हिन्दी" ,   # Hindi
        "hif": "Fiji Hindi",
        "ho": "Hiri Mot",        # Hiri Motu
        "hr": "Hrvatski" ,       # Croatian
        "hsb": "Hornjoserbsce",
        "ht": "Krèyol ayisyen" , # Haitian; Haitian Creole
        "hu": "Magyar" ,         # Hungarian
        "hy": "Հայերեն",         # Armenian
        "hz": "Otsiherero",      # Herero

        "ia": "Interlingua",
        "ie": "Interlingue",
        "id": "Bahasa Indonesia",# Indonesian (formerly in)
        "ig": "Igbo",            # Igbo
        "ii": "ꆇꉙ     ",       # Sichuan Yi
        "ik": "Iñupiak",         # Inupiak
        "ilo": "Ilokano",
        "io": "Ido" ,
        "is": "Íslenska" ,       # Icelandic
        "it": "Italiano" ,       # Italian
        "i": "ᐃᓄᒃᑎᑐᑦ",           # Inuktitut

        "ja": "日本語",          # Japanese
        "jbo": "Lojban",
        "jv": "Basa Jawa",       # Javanese

        "ka": "ქართული" ,        # Georgian
        "kaa": "Qaraqalpaqsha",
        "kab": "Taqbaylit",
        "kg": "KiKongo",         # Kongo
        "ki": "Gĩkũyũ",          # Kikuyu
        "kj": "Kuanyama",        # Kuanyama
        "kk": "Қазақша",         # Kazakh
        "kl": "Kalaallisut",     # Kalaallisut; Greenlandic
        "km": "ភាសាខ្មែរ",       # Khmer; Cambodian
        "kn": "ಕನ್ನಡ",           # Kannada
        "ko": "한국어" ,         # Korean
        "kr": "Kanuri",          # Kanuri
        "ks": "कश्मीरी / كشميري", # Kashmiri
        "ksh": "Ripoarisch",
        "ku": "Kurdî / كوردی",   # Kurdish
        "kv": "Коми",            # Komi
        "kw": "Kernewek/Karnuack", # Cornish
        "ky": "Кыргызча",        # Kirghiz

        "la": "Latina" ,         # Latin
        "lad": "Dzhudezmo",
        "lb": "Lëtzebuergesch" , # Letzeburgesch
        "lbe": "Лакку",
        "lg": "Luganda",         # Ganda
        "li": "Limburgs",        # Limburgish; Limburger; Limburgan
        "lij": "Lígur",
        "ln": "Lingala",         # Lingala
        "lmo": "Lumbaart",
        "lo": "ລາວ",             # Lao; Laotian
        "lt": "Lietuvių" ,       # Lithuanian
        "lua": "Luba",           # Luba
        "lv": "Latvieš" ,        # Latvian; Lettish

        "map-bms": "Basa Banyumasan",
        "mdf": "Мокшень (Mokshanj Kälj)",
        "mg": "Malagasy",        # Malagasy
        "mh": "Ebon",            # Marshall
        "mi": "Māori",           # Maori
        "mk": "Македонски" ,     # Macedonian
        "ml": None,              # Malayalam
        "mn": "Монгол",          # Mongolian
        "mo": "Молдовеняскэ",    # Moldavian
        "mr": "मराठी" ,     # Marathi
        "ms": "Bahasa Melay" ,   # Malay
        "mt": "Malti",           # Maltese
        "mus": "Muskogee",
        "my": "မ္ရန္‌မာစာ",      # Burmese
        "myv": "Эрзянь (Erzjanj Kelj)",
        "mzn": "مَزِروني",

        "na": "dorerin Naoero",  # Nauru
        "nah": "Nāhuatl",
        "nap": "Nnapulitano",
        "nb": "Norsk (Bokmål)",  # Norwegian Bokm@aa{}l
        "nd": None,              # Ndebele, North
        "nds": "Plattdüütsch",
        "nds-nl": "Nedersaksisch",
        "ne": "नेपाली",    # Nepali
        "new": "नेपाल भाषा" , # Nepal Bhasa
        "ng": "Oshiwambo",       # Ndonga
        "nl": "Nederlands" ,     # Dutch
        "nn": "Nynorsk",         # Norwegian Nynorsk
        "no": "Norsk (Bokmål)" , # Norwegian
        "nov": "Novial",
        "nr": None,              # Ndebele, South
        "nrm": "Nouormand/Normaund",
        "nv": "Diné bizaad",     # Navajo
        "ny": "Chi-Chewa",       # Chichewa; Nyanja

        "oc": "Occitan",         # Occitan; Proven@,{c}al
        "oj": None,              # Ojibwa
        "om": "Oromoo",          # (Afan) Oromo
        "or": "ଓଡ଼ିଆ",           # Oriya
        "os": "Иронау",          # Ossetian; Ossetic

        "pa": "ਪੰਜਾਬੀ" ,         # Panjabi; Punjabi
        "pag": "Pangasinan",
        "pam": "Kapampangan",
        "pap": "Papiament",
        "pdc": "Deitsch",
        "pi": "पाऴि",        # Pali
        "pih": "Norfuk",
        "pl": "Polski" ,         # Polish
        "pms": "Piemontèis" ,
        "ps": "پښتو",            # Pashto, Pushto
        "pt": "Português" ,      # Portuguese

        "q": "Runa Simi" ,       # Quechua

        "rm": "Rumantsch",       # Rhaeto-Romance
        "rmy": "romani - रोमानी",
        "rn": "Kirundi",         # Rundi; Kirundi
        "ro": "Română" ,         # Romanian
        "roa-rup": "Armãneashce",
        "roa-tara": "Tarandíne",
        "ru": "Русский" ,        # Russian
        "rw": "Ikinyarwanda",    # Kinyarwanda

        "sa": "संस्कृतम्", # Sanskrit
        "sah": "Саха тыла (Saxa Tyla)",
        "sc": "Sardu",           # Sardinian
        "scn": "Sicilian",
        "sco": "Scots",
        "sd": "سنڌي، سندھی ، सिन्ध", # Sindhi
        "se": "Sámegiella",      # Northern Sami
        "sg": "Sängö",           # Sango; Sangro
        "sh": "Srpskohrvatski / Српскохрватски" ,
        "si": "සිංහල",
        "simple": "Simple English" ,
        "sk": "Slovenčina" ,     # Slovak
        "sl": "Slovenščina" ,    # Slovenian
        "sm": "Gagana Samoa",    # Samoan
        "sn": "chiShona",        # Shona
        "so": "Soomaaliga",      # Somali
        "sr": "Српски / Srpski", # Serbian
        "srn": "Sranantongo",
        "ss": "SiSwati",         # Swati; Siswati
        "st": "Sesotho",         # Sesotho; Sotho, Southern
        "stk": "Seeltersk",
        "s": "Basa Sunda",       # Sundanese
        "sq": "Shqip" ,          # Albanian
        "szl": "Ślůnski",
        "sv": "Svenska" ,        # Swedish
        "sw": "Kiswahili",       # Swahili

        "ta": "தமிழ்" ,          # Tamil
        "te": "తెలుగు" ,         # Telugu
        "tet": "Tetun",
        "tg": "Тоҷикӣ",          # Tajik
        "th": "ไทย" ,            # Thai
        "ti": "ትግርኛ",            # Tigrinya
        "tk": "تركمن / Туркмен", # Turkmen
        "tl": "Tagalog" ,        # Tagalog
        "tn": "Setswana",        # Tswana; Setswana
        "to": "faka Tonga",      # Tonga (?) # Also ZW ; MW
        "tokipona": "Tokipona",
        "tpi": "Tok Pisin",
        "tr": "Türkçe" ,         # Turkish
        "ts": "Xitsonga",        # Tsonga
        "tt": "Tatarça / Татарча", # Tatar
        "tum": "chiTumbuka",
        "tw": "Twi",             # Twi
        "ty": "Reo Mā`ohi",      # Tahitian

        "udm": "Удмурт кыл",
        "ug": "Oyghurque",       # Uighur
        "uk": "Українська" ,     # Ukrainian
        "ur": "اردو",            # Urdu
        "uz": "O‘zbek",          # Uzbek

        "ve": "Tshivenda",       # Venda
        "vec": "Vèneto",
        "vi": "Tiếng Việt" ,     # Vietnamese
        "vls": "West-Vlams",
        "vo": "Volapük" ,

        "wa": "Walon",           # Walloon
        "war": "Winaray",
        "wo": "Wolof",           # Wolof
        "w": "吴语",

        "xal": "Хальмг",
        "xh": "isiXhosa",        # Xhosa

        "yi": "ייִדיש",          # Yiddish
        "yo": "Yorùbá",          # Yoruba

        "za": "Cuengh",          # Zhuang
        "zea": "Zeêuws",
        "zh": "中文" ,           # Chinese
        "zh-classical": "古文 / 文言文",
        "zm-min-nan": "Bân-lâm-gú",
        "zh-yue": "粵語",
        "zu": "isiZulu"          # Zulu
    }



# Wiki tokens. -*- coding: utf-8 -*-
# Copyright (C) 2015-2018 Sergey Poznyakoff
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Wiki markup tokens and associated classes.

This module defines classes for the basic nodes of the Wiki markup parse tree:

WikiNode        -- Abstract parse tree node.
WikiContentNode -- A node associated with some content.
WikiSeqNode     -- A sequence of nodes.
WikiTextNode    -- Textual content.
WikiDelimNode   -- Delimiter.
WikiTagNode     -- Tag (e.g. <tt>, </tt>, <tt />, etc.)
WikiRefNode     -- Wiki reference (e.g. [target|name])
WikiHdrNode     -- Heading (e.g. == Section ==)
WikiEltNode     -- Environment element.
WikiEnvNode     -- Environment (numbered or unnumbered list, definition, etc.)
WikiIndNode     -- Indent node.

Auxiliary classes:

WikiNodeEncoder -- Custom JSONEncoder subclass for serializing objects of the
                   above classes.
"""

from __future__ import print_function
import re
import json


class WikiNodeEncoder(json.JSONEncoder):
    """Custom JSONEncoder subclass for serializing WikiNode and its subclasses."""

    def default(self, obj):
        if isinstance(obj, WikiNode):
            return obj.json_encode()
        return json.JSONEncoder.default(self, obj)


def jsonencoder(func):
    def _mkencoder(self):
        json = func(self)
        json['wikinode'] = self.__class__.__name__
        json['type'] = self.type
        return json

    return _mkencoder


class WikiNode(object):
    """Generic parse tree node.

    Attributes:

      type   -- actual type of this object (string)
      parser -- parser instance that owns this node
    """

    type = 'UNDEF'
    parser = None

    def __init__(self, parser, **kwargs):
        self.parser = parser
        for key in kwargs:
            if hasattr(self, key):
                self.__dict__[key] = kwargs[key]
            else:
                raise AttributeError("'%s' has no attribute '%s'" % (self.__class__.__name__, key))

    def __str__(self):
        return json.dumps(self, cls=WikiNodeEncoder, sort_keys=True)

    @jsonencoder
    def json_encode(self):
        ret = {}
        for x in dir(self):
            if x == 'parser' or x.startswith('_') or type(x) == 'function':
                continue
            if x in self.__dict__:
                ret[x] = self.__dict__[x]
        return ret

    def format(self):
        """Abstract formatting function.

           Derived classes must override it.
        """
        pass


class WikiContentNode(WikiNode):
    """Generic content node.

    Attributes:

    content   --   Actual content
    """

    content = None

    def format(self):
        pass

    @jsonencoder
    def json_encode(self):
        ret = {}
        if self.content:
            if self.type == 'TEXT':
                ret['content'] = self.content
            elif isinstance(self.content, list):
                ret['content'] = [x for x in
                                  map(lambda x: x.json_encode(), self.content)]
            elif isinstance(self.content, WikiNode):
                ret['content'] = self.content.json_encode()
            else:
                ret['content'] = self.content
        else:
            ret['content'] = None
        return ret


class WikiSeqNode(WikiContentNode):
    """Generic sequence of nodes.

    Attributes:

    content  -- list of nodes.
    """

    def format(self):
        for x in self.content:
            x.format()

    @jsonencoder
    def json_encode(self):
        ret = {}
        if not self.content:
            ret['content'] = None
        elif isinstance(self.content, list):
            ret['content'] = [x for x in map(lambda x: x.json_encode(), self.content)]
        elif isinstance(self.content, WikiNode):
            ret['content'] = self.content.json_encode()
        else:
            ret['content'] = self.content
        return ret


# ##############

class WikiTextNode(WikiContentNode):
    """Text node.

    Attributes:

    type     -- 'TEXT'
    content  -- string
    """

    type = 'TEXT'

    @jsonencoder
    def json_encode(self):
        return {
            'content': self.content
        }


class WikiDelimNode(WikiContentNode):
    """Delimiter node.

    Attributes:

    type         -- 'DELIM'
    content      -- actual delimiter string
    isblock      -- boolean indicating whether it is a block delimiter
    continuation -- True if continuation is expected
    """

    type = 'DELIM'
    isblock=False
    continuation = False


class WikiTagNode(WikiContentNode):
    """A Wiki tag.

    Attributes:

    tag      -- actual tag name (with '<', '>', and eventual '/' stripped)
    isblock  -- True if this is a block tag
    args     -- List of tag arguments
    idx      -- If this is a "see also" reference, index of this ref in the
                list of references.
                FIXME: Perhaps this merits a subclass?
    """

    tag = None
    isblock = False
    args = None
    idx = None

    def __init__(self, *args, **keywords):
        super(WikiTagNode, self).__init__(*args, **keywords)
        if (self.type == 'TAG'
            and self.tag == 'ref'
            and hasattr(self.parser, 'references')):
            self.idx = len(self.parser.references)
            self.parser.references.append(self)

    @jsonencoder
    def json_encode(self):
        return {
            'tag': self.tag,
            'isblock': self.isblock,
            'args': self.args.tab if self.args else None,
            'content': self.content.json_encode() if self.content else None,
            'idx': self.idx
        }


class WikiRefNode(WikiContentNode):
    """Reference node.

    This class represents a wiki reference, such as "[ref|content]".

    Attributes:

    ref        -- actual reference
    content    -- content string
    """

    type = 'REF'
    ref = None
    @jsonencoder
    def json_encode(self):
        return {
            'ref': self.ref,
            'content': self.content.json_encode()
        }


class WikiHdrNode(WikiContentNode):
    """A wiki markup header class.

    Attributes:

    level    --  header level
    content  --  header content (WikiNode subclass object)
    """

    type = 'HDR'
    level = None

    @jsonencoder
    def json_encode(self):
        return {
            'level': self.level,
            'content': self.content.json_encode()
        }


class WikiEltNode(WikiContentNode):
    """Environment element node.

    Attributes:

    subtype    -- type of the environment (numbered, unnumbered, defn)
    content    -- content of the element (WikiNode subclass object)
    """

    type = 'ELT'
    subtype = None

    @jsonencoder
    def json_encode(self):
        return {
            'subtype': self.subtype,
            'content': self.content.json_encode()
        }


class WikiEnvNode(WikiContentNode):
    """Wiki Environment Node

    Attributes:

    envtype    -- type of the environment (numbered, unnumbered, defn)
    level      -- nesting level of the environment
    """

    type = 'ENV'
    envtype = None
    level = None

    @jsonencoder
    def json_encode(self):
        return {
            'envtype': self.envtype,
            'level': self.level,
            'content': [x for x in map(lambda x: x.json_encode(), self.content)]
        }


class WikiIndNode(WikiContentNode):
    """Indented block node.

    Attributes:

    level   -- Indentation level.
    content -- Indented content (WikiNode subclass object).
    """

    type = 'IND'
    level = None

    @jsonencoder
    def json_encode(self):
        return {
            'level': self.level,
            'content': self.content.json_encode()
        }



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (C) 2008-2018 Sergey Poznyakoff
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Wiki markup to HTML translator.

Classes:

HtmlWikiMarkup       -- Converts Wiki material to HTML.
HtmlWiktionaryMarkup -- Reserved for future use. Currently does the same as
                        HtmlWikiMarkup.

"""

from __future__ import print_function
from wikitrans.wikimarkup import *
from wikitrans.wikitoken import *
from wikitrans.wikins import wiki_ns_re, wiki_ns
import re
try:
    from urllib import quote as url_quote
except ImportError:
    from urllib.parse import quote as url_quote

try:
    from html import escape as html_escape
except ImportError:
    from cgi import escape as html_escape

__all__ = [ "HtmlWikiMarkup", "HtmlWiktionaryMarkup" ]


class HtmlSeqNode(WikiSeqNode):
    def format(self):
        s = ''
        for x in self.content:
            s += x.format()
        return s


class HtmlLinkNode(HtmlSeqNode):
    def format(self):
        arg = self.content[0].format()
        text = None
        if len(self.content) > 1:
            s = [x for x in map(lambda x: x.format(), self.content)]
            if s[0] == 'disambigR' or s[0] == 'wikiquote':
                return ""
            elif len(s) > 1 and s[1] == 'thumb':
                return ""
            text = '<span class="template">' + s[1] + '</span>'
            if self.type == 'TMPL':
                if re.match("t[+-]$", s[0]):
                    if len(s) > 2:
                        text = s[2]
                elif s[0] == "term":
                    text = self.parser.tmpl_term(s)
                elif s[0] == "proto":
                    text = self.parser.tmpl_proto(s)
                return text

        (qual, sep, tgt) = arg.partition(':')
        if tgt != '':
            ns = self.parser.wiki_ns_name(qual)
            if ns:
                if ns == 'NS_IMAGE':
                    return ''
                elif ns == 'NS_MEDIA':
                    tgt = self.parser.media_base + '/' + tgt
                else:
                    tgt = self.parser.mktgt(tgt)
            elif self.type == 'LINK' and qual in self.parser.langtab:
                tgt = self.parser.mktgt(tgt, qual)
                if not text or text == '':
                    text = self.parser.langtab[qual]
            else:
                tgt = self.parser.mktgt(tgt)
        else:
            tgt = self.parser.mktgt(arg)
        return "<a href=\"%s\">%s</a>" % (tgt,
                                          text if (text and text != '') else arg)


class HtmlRefNode(WikiRefNode):
    def format(self):
        target = self.ref
        text = self.content.format()
        return "<a href=\"%s\">%s</a>" % (
            target,
            text if (text and text != '') else target
        )


class HtmlFontNode(HtmlSeqNode):
    def format(self):
        comm = { 'IT': 'i',
                 'BOLD': 'b' }
        s = '<%s>' % comm[self.type]
        for x in self.content:
            s += x.format()
        s += '</%s>' % comm[self.type]
        return s


class HtmlTextNode(HtmlSeqNode):
    def format(self):
        if isinstance(self.content, list):
            s = ''.join(self.content)
        else:
            s = html_escape(self.content, quote=False)
        return s


class HtmlHdrNode(WikiHdrNode):
    def format(self):
        level = self.level
        if level > 6:
            level = 6
        return "<h%s>%s</h%s>\n\n" % (level, self.content.format(), level)


class HtmlBarNode(WikiNode):
    def format(self):
        return "<hr/>\n"


class HtmlEnvNode(WikiEnvNode):
    def format(self):
        type = self.envtype
        lev = self.level
        if lev > 4:
            lev = 2
        string = ""
        for s in self.content:
            n = s.subtype;
            string += "<%s>%s</%s>" % (self.parser.envt[type]["elt"][n],
                                       s.content.format(),
                                       self.parser.envt[type]["elt"][n])
        return "<%s>%s</%s>" % (self.parser.envt[type]["hdr"],
                                string,
                                self.parser.envt[type]["hdr"])
        return string


class HtmlTagNode(WikiTagNode):
    def format(self):
        if self.tag == 'code':
            self.parser.nested += 1
            s = self.content.format()
            self.parser.nested -= 1
            return '<pre><code>' + s + '</code></pre>' #FIXME
        elif self.tag == 'ref':
            n = self.idx+1
            return '<sup id="cite_ref-%d" class="reference"><a name="cite_ref-%d" href=#cite_note-%d">%d</a></sup>' % (n, n, n, n)
        elif self.tag == 'references':
            s = '<div class="references">\n'
            s += '<ol class="references">\n'
            n = 0
            for ref in self.parser.references:
                n += 1
                s += ('<li id="cite_note-%d">'
                      + '<span class="mw-cite-backlink">'
                      + '<b><a href="#cite_ref-%d">^</a></b>'
                      + '</span>'
                      + '<span class="reference-text">'
                      + ref.content.format()
                      + '</span>'
                      + '</li>\n') % (n, n)
            s += '</ol>\n</div>\n'
            return s
        else:
            s = '<' + self.tag
            if self.args:
                s += ' ' + str(self.args)
            s += '>'
            s += self.content.format()
            return s + '</' + self.tag + '>'


class HtmlParaNode(HtmlSeqNode):
    def format(self):
        return "<p>" + super(HtmlParaNode, self).format() + "</p>\n"


class HtmlPreNode(HtmlSeqNode):
    def format(self):
        s = super(HtmlPreNode, self).format()
        if self.parser.nested:
            return s
        else:
            return '<pre>' + s + '</pre>'


class HtmlIndNode(WikiIndNode):
    def format(self):
        return ("<dl><dd>" * self.level) + self.content.format() + "</dd></dl>" * self.level


class HtmlWikiMarkup(WikiMarkup):
    """A Wiki markup to HTML translator class.

    Usage:

      x = HtmlWikiMarkup(file="input.wiki")
      # Parse the input:
      x.parse()
      # Print it as HTML:
      print(str(x))

    Known bugs:
      * [[official position]]s
       Final 's' gets after closing </a> tag. Should be before.
    """

    nested = 0
    references = []
    def __init__(self, *args, **kwargs):
        """Create a HtmlWikiMarkup object.

        HtmlWikiMarkup([filename=FILE],[file=FD],[text=STRING],[lang=CODE],
                       [html_base=URL],[image_base=URL],[media_base=URL])

        The arguments have the same meaning as in the WikiMarkup constructor.

        """

        super(HtmlWikiMarkup, self).__init__(*args, **kwargs)
        self.token_class['LINK'] = HtmlLinkNode
        self.token_class['TMPL'] = HtmlLinkNode
        self.token_class['REF'] = HtmlRefNode
        self.token_class['IT'] = HtmlFontNode
        self.token_class['BOLD'] = HtmlFontNode
        self.token_class['HDR'] = HtmlHdrNode
        self.token_class['BAR'] = HtmlBarNode
        self.token_class['ENV'] = HtmlEnvNode
        self.token_class['TAG'] = HtmlTagNode
        self.token_class['PARA'] = HtmlParaNode
        self.token_class['PRE'] = HtmlPreNode
        self.token_class['IND'] = HtmlIndNode
        self.token_class['TEXT'] = HtmlTextNode
        self.token_class['SEQ'] = HtmlSeqNode

    def wiki_ns_name(self, str):
        if str in wiki_ns[self.lang]:
            return wiki_ns[self.lang][str]
        elif str in wiki_ns_re[self.lang]:
            for elt in wiki_ns_re[self.lang][str]:
                if str.beginswith(elt[0]) and str.endswith(elt[1]):
                    return elt[2]
        return None

    envt = { "unnumbered": { "hdr": "ul",
                             "elt": ["li"] },
             "numbered":   { "hdr": "ol",
                             "elt": ["li"] },
             "defn":       { "hdr": "dl",
                             "elt": ["dt","dd"] } }

    def mktgt(self, tgt, lang = None):
        if not lang:
            lang = self.lang
        return self.html_base % { 'lang' : lang } + url_quote(tgt)

    def tmpl_term(self, s):
        if len(s) == 2:
            return s[1]
        text = None
        trans = None
        for x in s[1:]:
            m = re.match('(\w+)=', x)
            if m:
                if m.group(1) == "tr":
                    trans = x[m.end(1)+1:]
            elif not text:
                text = x
        if text:
            if trans:
                text += ' <span class="trans">[' + trans + ']</span>'
        return text

    def tmpl_proto(self, s):
        text = '<span class="proto-lang">Proto-' + s[1] + '</span>'
        if len(s) >= 4:
            n = 0
            for x in s[2:-2]:
                if n > 0:
                    text += ','
                n += 1
                text += ' <span class="proto">' + x + '</span>'
                text += ' <span class="meaning">(' + s[-2] + ')</span>'
        return text

    def __str__(self):
        str = ""
        for elt in self.tree:
            str += elt.format()
        return str


class HtmlWiktionaryMarkup(HtmlWikiMarkup):
    """A class for translating Wiktionary articles into HTML.

    Reserved for future use. Currently does the same as HtmlWikiMarkup.
    """

    html_base='http://%(lang)s.wiktionary.org/wiki/'



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import print_function
import unittest
from wikitrans.wiki2text  import TextWikiMarkup
from wikitest import populate_methods

class TestTextWikiMarkup (unittest.TestCase):
    pass

populate_methods(TestTextWikiMarkup, TextWikiMarkup, '.text')

if __name__ == '__main__':
    unittest.main()



# Wiki "dump" format. -*- coding: utf-8 -*-
# Copyright (C) 2015-2018 Sergey Poznyakoff
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Print Wiki parse tree as JSON.

Classes:

DumpWikiMarkup

"""

from __future__ import print_function
from wikitrans.wikitoken import *
import json
from wikitrans.wikimarkup import WikiMarkup


class DumpReferences(object):
    idx = 0
    def __len__(self):
        return self.idx + 1
    def append(self, obj):
        self.idx += 1

class DumpWikiMarkup(WikiMarkup):
    """Produce a JSON dump of the Wiki markup parse tree.

    Usage:

      x = DumpWikiMarkup(file="input.wiki")
      # Parse the input:
      x.parse()
      # Print a JSON dump of the parse tree
      print(str(x))

    """

    indent = None
    references = DumpReferences()

    def __init__(self, **kwarg):
        """Create a DumpWikiMarkup object.

        Arguments:

        filename=FILE
          Read Wiki material from the file named FILE.
        file=FD
          Read Wiki material from file object FD.
        text=STRING
          Read Wiki material from STRING.
        indent=N
          Basic indent offset for JSON objects.
        """
        n = kwarg.pop('indent', None)
        if n != None:
            self.indent = int(n)
        super(DumpWikiMarkup, self).__init__(self, **kwarg)

    def __str__(self):
        return json.dumps(self.tree,
                          cls=WikiNodeEncoder,
                          indent=self.indent,
                          separators=(',', ': '),
                          sort_keys=True)



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import print_function
from glob import glob
import os.path

def wiki_markup_test(classname, name_in, name_out):
    fh = open(name_out)
    buf = ''.join(fh.readlines()).strip()
    fh.close()
    hwm = classname(filename=name_in, lang="en")
    hwm.parse()

    if str(hwm).strip() == buf:
        return True

    # fail
    print("\n>>>%s<<<" % buf)
    print(">>>%s<<<" % str(hwm).strip())
    return False

def populate_methods(cls, wcls, suffix):
    def settest(self, base, wiki_name, pat_name):
        def dyntest(self):
            self.assertTrue(wiki_markup_test(wcls, wiki_name, pat_name))
        meth = 'test_' + wcls.__name__ + '_' + base
        dyntest.__name__ = meth
        setattr(cls, meth, dyntest)
    for file in glob('testdata/*.wiki'):
        if os.path.isfile(file):
            patfile = file[:len(file) - 5] + suffix
            base, ext = os.path.splitext(os.path.basename(file))
            if os.path.exists(patfile) and os.path.isfile(patfile):
                settest(cls, base, file, patfile)



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import print_function
import unittest
from wikitrans.wiki2texi  import TexiWikiMarkup
from wikitest import populate_methods

class TestTexiWikiMarkup (unittest.TestCase):
    pass

populate_methods(TestTexiWikiMarkup, TexiWikiMarkup, '.texi')

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import print_function
import unittest
from wikitrans.wiki2html  import HtmlWikiMarkup
from wikitest import populate_methods

class TestWikiMarkup (unittest.TestCase):
    pass

populate_methods(TestWikiMarkup, HtmlWikiMarkup, '.html')

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2008-2018 Sergey Poznyakoff
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup, find_packages
from codecs import open

with open("README.rst", "r") as fh:
    long_description = fh.read()

setup(name='wikitrans',
      version='1.3',
      author='Sergey Poznyakoff',
      author_email='gray@gnu.org',
      url='http://www.gnu.org.ua/projects/wikitrans',
      packages = find_packages(exclude=['contrib', 'docs',
                                        'tests', 'testdata']),
      scripts=['bin/wikitrans'],
      license='GPL License',
      description='Wiki markup translator.',
      long_description=long_description,
      long_description_content_type="text/x-rst",
      platforms=['any'],
      test_suite='tests',
      classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Text Processing :: Markup'
      ],
      keywords = 'mediawiki markup translation',
)
